# Copyright (c) 2022 - for information on the respective copyright owner see the NOTICE.rst file or
# the repository https://github.com/boschglobal/luisy
#
# SPDX-License-Identifier: Apache-2.0
"""
This module contains all the management of luisy's configuration. It holds a singleton, that can
be used anywhere in the project after it was initialized in :py:mod:`luisy.__init__`. Also the
cli parameters are passes into this config file in py:func:`luisy.cli.luisy_run()`.
This singleton allows us to access parameters like `working_dir`, `download`, ... anywhere in our
pipelines. We don't need to pass arguments through that pipeline anymore to get the information
into the leafs of our DAG.
"""

import os
import logging
from luisy.file_system import AzureContainer
from luisy.default_params import (
    default_params,
    env_keys
)

logger = logging.getLogger('luigi-interface').getChild('luisy-interface')


def get_default_params(raw=True):
    if raw:
        return default_params.copy()
    else:
        return {key: val for key, val in default_params.items() if val is not None}


class Singleton(type):
    _instances = {}

    def __call__(cls, *args, **kwargs):
        if cls not in cls._instances:
            cls._instances[cls] = super(Singleton, cls).__call__(*args, **kwargs)
        return cls._instances[cls]


class Config(metaclass=Singleton):

    def __init__(self):
        self._config = get_default_params()

        self._files_to_download = []

        for param, env_key in env_keys.items():
            self.set_param(param, self._get_env_var(env_key))

        if self.get_param('azure_storage_key') is not None:
            self.fs = AzureContainer(
                account_name=self.get_param('azure_account_name'),
                container_name=self.get_param('azure_container_name'),
                key=self.get_param('azure_storage_key')
            )

    def update(self, params):
        """
        Takes dict and updates the config with all entries in that dict

        Args:
            params (dict): new params that should be set
        """
        self._config.update(params)

    def set_param(self, name, val):
        """
        Set param in config.

        Args:
            name (str): Key to parameter
            val (object): Value of parameter
        """
        self._config[name] = val

    def get_param(self, param):
        """
        Get param from config.

        Args:
            param (str): Key to parameter

        Returns:
            Value of parameter
        """
        if param not in self._config:
            raise ValueError(f'{param} not found in Config. Please add to default params')
        return self._config[param]

    @property
    def download(self):
        return self.get_param('download')

    @property
    def upload(self):
        return self.get_param('upload')

    @property
    def working_dir(self):
        return os.path.normpath(self.get_param('working_dir'))

    @property
    def config(self):
        """
        Get the whole config

        Returns:
            dict: Config with all parameter values set right now
        """
        return self._config

    def _get_env_var(self, param):
        """
        Tries to get given param out of the environment variables. The logger throws a warning if
        the variable cannot be found in the system

        Args:
            param(str): key of environ variable e.g. `WORKING_DIR`

        Returns:
            str or None: Value of environ variable
        """
        if param not in os.environ:
            logger.warning(f'Environment Variable {param} not set!')
            return None
        return os.environ[param]

    def reset(self):
        """
        Resets the config singleton to the initial state with default parameters and environment
        variable values.
        """
        self._config = None
        self.__init__()

    def check_params(self):
        """
        Way to check if parameters are valid and luisy is ready to initiate the luigi run.
        Currently `working_dir` has to be set and also if the user wants to use the cloud, luisy
        only allows runs when `azure_storage_key`, `azure_account_name`, and `azure_container_name`
        are set.

        Raises:
            ValueError: When parameters are wrong or missing

        """
        needs_azure = self._config['download'] or self._config['upload']

        for azure_param in ['azure_storage_key', 'azure_container_name', 'azure_account_name']:

            if (self._config[azure_param] is None) and needs_azure:
                raise ValueError(
                    f"Parameter '{azure_param}' not set. You cant use download "
                    "or upload functionality without setting your Azure storage key. More "
                    "information: docs/cloud.rst"
                )
        if self._config['working_dir'] is None:
            raise ValueError("Parameter 'working_dir' not set!")


def pass_args(args):
    Config().update(args)


def activate_download():
    Config().set_param('download', True)


def set_working_dir(working_dir):
    Config().set_param('working_dir', working_dir)


def remove_working_dir(path):
    """
    Removes the working dir from the filepath.

    Example:

        If :code:`path=/data/my_project/raw/some_file.pkl`, then the output is
        :code:`/my_project/raw/some_file.pkl`

    Args:

        path (str): Path to a file

    Returns:
        str: Path to the file where working dir has been removed.
    """
    return change_working_dir(
        path=path,
        dir_current=Config().working_dir,
        dir_new='')


def add_working_dir(path):
    """
    Adds the working dir to a filepath.

    Example:

        If the input is :code:`path=/my_project/raw/some_file.pkl` and the working dir is
        :code:`/mnt/d/data`, then the output is :code:`/mnt/d/data/my_project/raw/some_file.pkl`.

    Args:

        path (str): Path to a file without working dir

    Returns:
        str: Path to the file with working dir prepended.
    """

    return os.path.join(Config().working_dir, path.lstrip('/'))


def change_working_dir(path, dir_current, dir_new):
    """
    Exchanges the working dir in :code:`path`

    Args:
        path (str): Path where working dir should be changed
        dir_current (str): The path of the working dir to be replaced
        dir_new (str): The working dir that should be inserted

    Returns:
        str: Path with updated working dir
    """

    # Make sure they do not end with slash
    dir_current = os.path.normpath(dir_current)
    assert path[:len(dir_current)] == dir_current

    if len(dir_new) > 0:
        dir_new = os.path.normpath(dir_new)
        return os.path.join(
            dir_new, path[len(dir_current):].lstrip('/')
        )
    else:
        return path[len(dir_current):]



# Copyright (c) 2022 - for information on the respective copyright owner see the NOTICE.rst file or
# the repository https://github.com/boschglobal/luisy
#
# SPDX-License-Identifier: Apache-2.0

import os
from azure.storage.blob import BlobServiceClient
import tempfile
import json
import logging
import shutil
from luisy.helpers import (
    get_df_from_parquet_dir,
    remove_parent_directories,
)
logger = logging.getLogger(__name__)


class AzureContainer(object):
    """
    Abstraction for an Azure container.

    Args:
        account_name (str): Account name of the azure storage
        container_name (str): Account name of the azure container
        key (str): Secret key of the container within the storage account
    """

    def __init__(self, account_name="", key=None, container_name=""):

        if key is None:
            raise ValueError('Key needs to be given')

        os.environ["AZURE_CLI_DISABLE_CONNECTION_VERIFICATION"] = "1"
        os.environ["ADAL_PYTHON_SSL_NO_VERIFY"] = "1"

        client = BlobServiceClient(
            account_url="https://{}.blob.core.windows.net".format(account_name),
            credential=key)
        self.client = client.get_container_client(container_name)

    def exists(self, path):
        """
        Checks if dir exists.

        Args:
            path (string): Path within the container, like :code:`/path/to/folder`

        Returns:
            bool: True, if folder exists. False otherwise.
        """
        path = self._adapt_remote_path(path)

        if path.endswith('/'):
            path = path[:-1]

        for blob in self.client.list_blobs(name_starts_with=path):
            # Either path is a directory, in this case the remote blob directory needs one
            # containing file and the path is then a substring of the blob name
            if blob['name'].startswith(path + '/'):
                return True

            # If the path is a file, then must be a blob with the exact same name
            if path == blob['name']:
                return True
        return False

    def _adapt_remote_path(self, path):
        if path[0] == '/':
            path = path[1:]
        return path

    def download(self, source, dest):
        """
        Downloads the content of the remote directory to the local directory.

        Args:
            source (str): Path to the directory within the container, like :code:`/path/to/data`
            dest (str): Path to local directory.

        Raises:
            ValueError: If local folder exists already
        """
        source = self._adapt_remote_path(source)
        if not dest:
            raise Exception('A destination must be provided')

        tmpdir = tempfile.mkdtemp()

        if os.path.exists(dest):
            raise ValueError('Local dir {} exists already'.format(dest))

        if not source.endswith('/'):
            source += '/'
        if not dest.endswith('/'):
            dest += '/'

        blobs = self._ls_files(source, recursive=True)
        file_blobs = remove_parent_directories(lst_with_substring_elements=blobs)
        file_blobs = [source + blob for blob in file_blobs]

        for blob in file_blobs:
            blob_dest = os.path.join(tmpdir, os.path.relpath(blob, source))
            self._download_file(blob, blob_dest)

        shutil.move(tmpdir, dest)

    def _ls_files(self, path, recursive=False):

        path = self._adapt_remote_path(path)
        if not path == '' and not path.endswith('/'):
            path += '/'

        blob_iter = self.client.list_blobs(name_starts_with=path)
        files = []
        for blob in blob_iter:
            relative_path = os.path.relpath(blob.name, path)
            if recursive or '/' not in relative_path:
                files.append(relative_path)
        return files

    def _download_file(self, source, dest):
        """
        dest is a directory if ending with '/' or '.', otherwise it's a file
        """
        source = self._adapt_remote_path(source)

        if dest.endswith('/'):
            dest = os.path.join(dest, os.path.basename(source))

        logger.info(f'Downloading {source} to {dest}')
        os.makedirs(os.path.dirname(dest), exist_ok=True)

        bc = self.client.get_blob_client(blob=source)
        with open(dest, 'wb') as file:
            data = bc.download_blob()
            file.write(data.readall())

    def _upload_file(self, source, dest, overwrite=False):
        """
        dest is a directory if ending with '/' or '.', otherwise it's a file
        """
        dest = self._adapt_remote_path(dest)
        if source.endswith('/'):
            source = os.path.join(source, os.path.basename(dest))

        bc = self.client.get_blob_client(blob=dest)
        with open(source, 'rb') as data:
            bc.upload_blob(data, overwrite=overwrite)

    def get_df(self, remote_path):
        """
        """
        remote_path = self._adapt_remote_path(remote_path)
        with tempfile.TemporaryDirectory() as tmp_dir:
            self.download(remote_path, tmp_dir)
            logger.info('Reading dataframe from tmpdir')
            df = get_df_from_parquet_dir(os.path.join(tmp_dir, os.path.basename(remote_path)))
        return df

    def save_json(self, filepath, data):

        bc = self.client.get_blob_client(blob=self._adapt_remote_path(filepath))
        bc.upload_blob(json.dumps(data), overwrite=True)

    def load_json(self, filepath):

        bc = self.client.get_blob_client(blob=self._adapt_remote_path(filepath))
        return json.loads(bc.download_blob().readall().decode())



# Copyright (c) 2022 - for information on the respective copyright owner see the NOTICE.rst file or
# the repository https://github.com/boschglobal/luisy
#
# SPDX-License-Identifier: Apache-2.0

import logging
import sys

log_format = '%(asctime)s,%(msecs)03d %(levelname).1s %(name)s %(message)s'

logging.basicConfig(
    stream=sys.stdout,
    format=log_format,
    level=logging.INFO)


from luisy.tasks import Task # noqa
from luisy.tasks import ExternalTask # noqa
from luisy.tasks import ConcatenationTask # noqa
from luisy.tasks import WrapperTask # noqa

from luisy.config import Config # noqa
from luisy.config import set_working_dir # noqa
from luisy.config import activate_download # noqa

from luisy.decorators import final # noqa
from luisy.decorators import interim # noqa
from luisy.decorators import raw # noqa

from luisy.decorators import inherits # noqa
from luisy.decorators import requires # noqa
from luisy.decorators import auto_filename # noqa

from luisy.decorators import csv_output # noqa
from luisy.decorators import json_output # noqa
from luisy.decorators import xlsx_output # noqa
from luisy.decorators import hdf_output # noqa
from luisy.decorators import pickle_output # noqa
from luisy.decorators import parquetdir_output # noqa

from luigi import Parameter # noqa
from luigi import IntParameter # noqa
from luigi import FloatParameter # noqa
from luigi import TupleParameter # noqa
from luigi import ListParameter # noqa
from luigi import DictParameter # noqa
from luigi import DateParameter # noqa
from luigi import DateIntervalParameter # noqa
from luigi import BoolParameter # noqa

Config()



# Copyright (c) 2022 - for information on the respective copyright owner see the NOTICE.rst file or
# the repository https://github.com/boschglobal/luisy
#
# SPDX-License-Identifier: Apache-2.0

import pickle
import json
import luigi
import pandas as pd
import os
import logging
import re

from luisy.config import (
    Config,
    change_working_dir,
)
from luisy.helpers import get_df_from_parquet_dir

logger = logging.getLogger(__name__)


class LocalTarget(luigi.LocalTarget):
    """

    Args:
        path (str): Path to the file
        working_dir (str): The working dir
        download (bool): Whether the file should be downloaded from the cloud if not available
            locally
    """

    file_ending = None

    def __init__(self, path, **kwargs):
        luigi.LocalTarget.__init__(
            self,
            path=path,
            format=None,
            is_tmp=False
        )
        self.kwargs = kwargs

    @property
    def fs(self):
        return Config().fs

    def is_folder(self):
        """
        Checks if the output is a folder.

        .. note::
            This may has to be improved in future as we (silently) assume that any file has file
            extension.

        """
        _, file_ending = os.path.splitext(os.path.basename(self.path))
        return file_ending == ''

    def make_dir(self, path):
        """
        Creates the local path `path`.

        """
        logger.info(f'Create dir {path}')
        os.makedirs(path, exist_ok=True)

    def remove(self):
        if os.path.exists(self.path):
            os.remove(self.path)
        else:
            logger.info(f'Target {self.path} not existing, nothing to remove')

    def exists(self):
        """
        Checks if the file exists. If :code:`download` is set to :code:`True`, then it is checked
        whether the file is available in the cloud if it is not available locally. If it is
        available, it is downloaded.
        """

        file_exists_locally = os.path.exists(self.path)

        if not file_exists_locally and Config().get_param('download'):
            if self.path in Config()._files_to_download:
                logger.info('Task not complete locally, check if file is available in cloud')
                self._try_to_download()

            # Check again if file exists
            file_exists_locally = os.path.exists(self.path)

        return file_exists_locally

    def _try_to_download(self):
        """
        Tries to download the target file from the cloud
        """
        path = change_working_dir(
            path=self.path,
            dir_current=Config().working_dir,
            dir_new='')

        if self.fs.exists(path):
            logger.info('File available. Start download')

            if self.is_folder():
                logger.info('Download folder')

                self.fs.download(
                    source=path,
                    dest=self.path)
            else:
                logger.info('Download file')
                self.fs._download_file(
                    source=path,
                    dest=self.path,
                )
        else:
            logger.info('File not in cloud.')

    def _try_to_upload(self, overwrite=False):
        """
        Tries to upload the target file to the cloud
        """
        path = change_working_dir(
            path=self.path,
            dir_current=Config().working_dir,
            dir_new='')

        if self.is_folder():
            raise NotImplementedError('Not implemented yet')

        logger.info('File not available. Start upload')
        self.fs._upload_file(
            source=self.path,
            dest=path,
            overwrite=overwrite,
        )


class PickleTarget(LocalTarget):
    file_ending = 'pkl'

    def write(self, obj):
        with open(self.path, 'wb') as f:
            pickle.dump(obj, f)

    def read(self):
        with open(self.path, 'rb') as f:
            return pickle.load(f)


class HDFTarget(LocalTarget):
    file_ending = 'hdf'

    def write(self, df):
        df.to_hdf(self.path, key='data')

    def read(self):
        return pd.read_hdf(self.path)


class XLSXTarget(LocalTarget):
    file_ending = 'xlsx'

    def write(self, df):
        df.to_excel(self.path)

    def read(self):
        return pd.read_excel(self.path, **self.kwargs)


class CSVTarget(LocalTarget):
    file_ending = 'csv'

    def write(self, df):
        df.to_csv(self.path, index=False, **self.kwargs)

    def read(self):
        return pd.read_csv(self.path, **self.kwargs)


class ParquetDirTarget(LocalTarget):
    file_ending = ''

    def write(self, df):
        raise NotImplementedError("Coming soon...")

    def read(self):
        return get_df_from_parquet_dir(self.path)


class JSONTarget(LocalTarget):
    file_ending = 'json'

    def write(self, dct):
        assert isinstance(dct, dict)
        with open(self.path, 'w') as f:
            json.dump(dct, f, indent=4, separators=(',', ': '))

    def read(self):
        with open(self.path, 'r') as f:
            return json.load(f)


class DirectoryTarget(LocalTarget):
    """
    Target to read directories. The read_file method has to be filled by the user and passed to
    target by decorating it.
    """
    file_ending = ''
    regex_pattern = None

    def __init__(self, path, **kwargs):
        luigi.LocalTarget.__init__(
            self,
            path=path,
            format=None,
            is_tmp=False
        )
        if self.regex_pattern is None:
            self.regex_pattern = ''

        self.kwargs = kwargs

    def write(self):
        raise NotImplementedError("Coming soon...")

    def is_valid_file(self, filename):
        """
        Checks if file matches given regex

        Args:
            filename (str): file that need to be checked against regex

        Returns:
            bool: flag if file matches regex

        """
        return re.search(self.regex_pattern, filename)

    def read(self):
        for i, filename in enumerate(os.listdir(self.path)):
            if self.is_valid_file(filename):
                content = self.read_file(os.path.join(self.path, filename))
                if content is not None:
                    yield content

    def read_file(self, file):
        raise NotImplementedError('Needs to be passed by user. See '
                                  'decorators.make_directory_output')



# Copyright (c) 2022 - for information on the respective copyright owner see the NOTICE.rst file or
# the repository https://github.com/boschglobal/luisy
#
# SPDX-License-Identifier: Apache-2.0

import os
import re
from functools import reduce
from luigi.date_interval import Week
from luigi.tools.deps import get_task_requires
import copy
from contextlib import contextmanager
import glob
import pandas as pd
import logging

logger = logging.getLogger(__name__)


def get_string_repr(obj):
    """
    Computes a string representation of the given object.

    Args:
        obj (object): An arbitrary python object

    Returns:
        str: A string representation
    """

    if hasattr(obj, '__iter__') and type(obj) != str:
        return "_".join([get_string_repr(o) for o in obj])
    else:
        return str(obj).replace(' ', '_').replace('/', '_')


def get_start_date(week):
    """
    Extracts the first day of the week.

    Args:
        week (luigi.date_interval.Week or string): Week that is compatible with luigi or string of
            the form :code:`2021-W12`

    Returns:
        datetime.date: The first day of the week
    """
    logger.info('passed in week is {}'.format(week))
    if isinstance(week, Week):
        logger.info('returning date is {}'.format(week.date_a))
        return week.date_a
    else:
        logger.info('returning date is {}'.format(Week.parse(week).date_a))
        return Week.parse(week).date_a


def get_stop_date(week):
    """
    Extracts the last day of the week.

    Args:
        week (luigi.date_interval.Week or string): Week that is compatible with luigi or string of
            the form :code:`2021-W12`

    Returns:
        datetime.date: The last day of the week
    """
    logger.info('passed in week is {}'.format(week))
    if isinstance(week, Week):
        logger.info('returning date is {}'.format(week.date_b))
        return week.date_b
    else:
        logger.info('returning date is {}'.format(Week.parse(week).date_b))
        return Week.parse(week).date_b


def get_all_reqs(task, visited=None):
    if visited is None:
        visited = [task]

    for next in get_task_requires(task) - set(visited):
        visited += get_all_reqs(next, visited=[next])

    return list(set(visited))


def add_filter_info(filter_func):
    """
    this decorator for class methods adds information about the change of a pd.DataFrame's shape
    before (and after if function returns a pd.DataFrame) filtering the df to the logger.
    This provides additional information to the user which function filtered how many rows / cols.
    This can be easily applied to every task method like this:

    .. code-block:: python

        @add_filter_info
        def filter_func(df):
            return df.drop(["my_most_disliked_column"], axis=1)

    Args:
        filter_func (cls.method): method that takes pd.DataFrame as first argument and filters it by
            some operations

    Returns:
        wrapper around filter_func, that adds additional information about filtered pd.DataFrame.

    """

    def wrapper(cls, df, *args, **kwargs):
        logger.info(f"{cls.__class__}:")
        logger.info(f"df.shape before '{filter_func.__name__}': {df.shape}")
        df = filter_func(cls, df, *args, **kwargs)
        if df is not None:
            logger.info(f"df.shape after '{filter_func.__name__}': {df.shape}")
            return df

    return wrapper


def wrap_in_brackets(string):
    return '(' + string + ')'


class RegexParam:
    """
    Class to hold information about a Parameter which RegexTaskPattern class uses to create
    matching regex queries.

    Args:
        name (str): Parameter name in luisy.Task
        default (object): parameters original default value
        regex_placeholder (str): regex string which matches all possible parameter values.
    """

    def __init__(self, name, default, regex_placeholder='[\\S]*'):
        self.name = name
        self.default = default
        self.regex_placeholder = regex_placeholder

    @property
    def placeholder(self):
        """
        This creates a unique placeholder, which is used for finding the location of the
        parameter inside a string. i.e. filepath.

        Returns:
            str: parameter surrounded with "*"
        """
        return f'*{self.name}*'


class RegexTaskPattern:
    """
    This class can be used to find instances of a task on the disk. It uses regex to search for
    in the user's filesystem. Each param is then replaced with a placeholder to build the regex
    from it.

    Args:
        task_cls (Type[luisy.Task]): a class object of type luisy.Task. The given task should use
            the :py:func:`luisy.decorator.auto_filename` to work properly.
        regex_placeholder (str): This is the regex which is inserted into the filename to create
            a matching  regex, which then searches the root directory.

    Note:
        Only works when using @auto_filename
    """

    def __init__(self, task_cls, regex_placeholder='[\\S]*'):
        self.task_cls = task_cls
        self.params = [
            RegexParam(
                name=param_name,
                default=copy.deepcopy(getattr(task_cls, param_name)._default),
                regex_placeholder=regex_placeholder
            ) for param_name in task_cls.get_param_names()
        ]
        self.regex_placeholder = regex_placeholder
        self.set_filename_pattern()

        self.regex_pattern = self.build_regex()
        self.root_dir = self.build_root_directory()

    @contextmanager
    def _overwrite_defaults(self):
        """
        This method works as a contextmanager to overwrite a task classes default params. This
        is used to extract filename strings, which can be used to build regex queries.

        Yields:
            Overwritten class where the parameter defaults are overwritten with placeholders

        """
        for param in self.params:
            getattr(self.task_cls, param.name)._default = param.placeholder
        yield self.task_cls
        for param in self.params:
            getattr(self.task_cls, param.name)._default = param.default

    def set_filename_pattern(self):
        """
        This creates a filename pattern with parameter values filled by
        :py:func:`RegexParam.placeholder()`
        """
        with self._overwrite_defaults() as task_cls:
            task = task_cls()
            self.filename_pattern = task.output().path

    def build_root_directory(self):
        """
        Finds "highest" possible path that can be used as a root directory.

        Returns:
            str: Root directory where the regex search is applied on the files inside it.

        """
        path, file = os.path.split(self.filename_pattern)
        dirs = path.split(os.sep)
        smallest_idx = None
        for param in self.params:
            if param.placeholder in dirs:
                smallest_idx = dirs.index(param.placeholder)
        if smallest_idx is not None:
            return os.sep.join(dirs[:smallest_idx])
        return path

    def build_regex(self):
        """
        Build a regex string from the filepath. This replaces the parameter values of the task
        with a regex pattern that matches these values.

        Returns:
            str: Regex pattern that can be used to search for files of that type

        """
        logger.info('Building Regex')
        regex_pattern = self.filename_pattern
        for param in self.params:
            regex_pattern = regex_pattern.replace(param.placeholder, param.regex_placeholder)
        path = os.path.normpath(regex_pattern)
        file_regex = wrap_in_brackets(self.regex_placeholder).join(
            [wrap_in_brackets(group) for group in path.split(
                self.regex_placeholder)]
        )

        return file_regex

    def build_re_splitters(self):
        """
        The splitters are needed to extract the parameters of a file that matches the regex.

        Returns:
            dict: keys are parameter names, values is a list of splits of the filesname. This
                allows us to separate the value from the filename

        """
        splits = {}
        for param in self.params:
            repls = [(p.placeholder, p.regex_placeholder) for p in self.params
                     if p.name != param.name]
            replaced_str = reduce(lambda a, kv: a.replace(*kv), repls, self.filename_pattern)
            splits[param.name] = replaced_str.split(param.placeholder)
        return splits

    def get_available_files(self):
        """
        Scans all files in the classes' root_directory and then filters on the regex_pattern of
        this class.

        Returns:
            list: List of files which match the regexpattern and are in the root directory

        """
        file_paths = []
        logger.info(f'Looking for matching regex in {self.root_dir}')
        for root, dirs, files in os.walk(self.root_dir):
            for file in files:
                file_paths.append(os.path.join(root, file))
        r = re.compile(self.regex_pattern)
        files = list(filter(r.match, file_paths))
        logger.info('Found matching files:')
        logger.info(files)
        return files

    def get_available_params(self):
        """
        Searches for Files in the rootdirectory matching to the regex and extracts the parameters
        out of the filesnames using the regex_splits.

        Returns:
            dict: keys are file_paths of the files found, values are a dict with param_name as
                key and param_value as value.

        """
        values = {}
        files = self.get_available_files()
        re_splitters = self.build_re_splitters()
        logger.info('Extract params from files')
        for file_path in files:
            file_params = {}
            for param in self.params:
                prefix_splitter = re_splitters[param.name][0]
                suffix_splitter = re_splitters[param.name][1]
                truncated_prefix = re.split(prefix_splitter, file_path)[1]
                file_params[param.name] = re.split(suffix_splitter, truncated_prefix)[0]
            values[file_path] = file_params
        logger.info('Params extracted:')
        logger.info(list(values.values()))
        return values


def get_related_instances(task_cls, regex_placeholder='[\\S]*'):
    regex_pattern = RegexTaskPattern(
        task_cls=task_cls,
        regex_placeholder=regex_placeholder
    )
    params = regex_pattern.get_available_params()
    return [task_cls(**entry) for entry in params.values()]


def get_df_from_parquet_dir(path):
    """
    Reads all Parquet files from a given directory and parses them into a pandas dataframe

    Args:
        path (str): path where parquet files are stored

    Returns:
        pd.DataFrame: dataframe with all the data from parquet files
    """

    files = glob.glob(os.path.join(path, '*.parquet'))
    dfs = [pd.read_parquet(f, engine='pyarrow') for f in files]

    # Remove all data frames with no lines
    dfs = filter(lambda df: df.shape[0] > 0, dfs)

    try:
        df = pd.concat(dfs)
        df.reset_index(inplace=True, drop=True)
    except ValueError:
        raise ValueError(f"Could not find any parquet files in '{path}'")

    return df


def remove_parent_directories(lst_with_substring_elements):
    """
    Removes elements from list that are substrings from other items in that list. E.g.
    ['/mnt/d', '/mnt/d/a', '/mnt/d/b'] removes '/mnt/d' from the list.
    Args:
        lst_with_substring_elements(list): removes substrings from that list

    Returns:
        List: Clean list without substrings of other elements, sorted by character length
    """
    lst_with_substring_elements.sort(key=lambda s: len(s), reverse=True)
    clean_lst = []
    for element in lst_with_substring_elements:
        if not any([element in file for file in clean_lst]):
            clean_lst.append(element)
    return clean_lst



from setuptools import (
    setup,
    find_packages,
)
from pathlib import Path


with open('requirements.txt') as f:
    requirements_install = f.read().splitlines()

with open('VERSION') as f:
    version = f.read().splitlines()[0]

setup(
    name='luisy',
    version=version,
    description='Framework to build data pipelines',
    author='Robert Bosch GmbH',
    entry_points={
        'console_scripts': [
            'luisy = luisy.cli:luisy_run',
        ]
    },
    packages=find_packages(exclude=['contrib', 'docs', 'tests']),
    install_requires=requirements_install,
    tests_require=[],
    license="Apache-2.0",
    long_description=(Path(__file__).parent / "README.md").read_text(),
    long_description_content_type='text/markdown',
    extras_require={
        'dev': ['check-manifest'],
        'test': ['coverage'],
    },
    project_urls={}
)



# Copyright (c) 2022 - for information on the respective copyright owner see the NOTICE.rst file or
# the repository https://github.com/boschglobal/luisy
#
# SPDX-License-Identifier: Apache-2.0

from luisy.cli import luisy_run

if __name__ == '__main__':
    luisy_run()
